#ifndef _DEFS_H_
#define _DEFS_H_

#include "config.h"
#include "error.h"
/* #include "spinsdef.h" 2015-12 removed */

#ifndef MAX_ITR		/* can be set now in config.h, Apr08 */
#define MAX_ITR  1000	/* maximum number of iterations, old=150 */
#endif

#if defined(__STRICT_ANSI__) || !(defined(__cplusplus) || defined(__GNUC__) || defined(__CPP__))
#define inline
#endif

/* please define VecType in config.h or Makefile DFLAGS !!! Mar99 JS */
#ifndef VecType     /* 2020-11 +1 = flt64 (double), ~1 = flt32 */
#if defined(__cplusplus)
#define VecType      5  /* 4 = complex float32 as default for C++ */
#else
#define VecType      1	/* ~4 =   real float32 as default for C */
#endif
/*
    ToDo: reorder VecTypes bit0,1:0-3=flt8,flt16,flt32,flt64, bit2:4=cplx
                  SumTypes 0=flt64,1=flt80-128
    0=float,1=double,8=halfdouble=flt16(16bit,untested)
    4=complex-flt, 5=cplx-dbl, 12=cplx-flt16 (gcc, c++ or c99) for huge arrays
    2=long double (for sums)
 */
#endif

/*  plan: use short only for large vectors (less memory)!
 s=small l=long (2013-04-04)
         VecType:           0          1        2         !4      
 typical | typedef  |  short-  | normal- | long-  | real-version | old
 --------+----------+----------+---------+--------+--------------+--------
 array     scplx     sComplex    Complex   Complex  sdouble        mcplx,mzahl
 sums      lcplx      Complex    Complex  lComplex  ldouble        cplx,zahl
 rest       cplx      Complex    Complex   Complex   double        cplx,double

 [sl]cplx is identical to [sl]double in real mode (only real EVs)

*/

#if   (VecType & (1+8)) == 8          /* __GNUC__ 2.5.8 */
/* 2013-04-04 16bit fix point float removed, bugs + bad 32bit tests
 * see ieee754.h union ieee754_float ... struct ...int ...:.. 32bit++
 * gcc-4.1.2 does not now about float16,
 *   ToDo:  does someone know about float16/binary16 defined in C99?
 * tested: 2019-06-06 _Float16 ... (~gcc/g++8.2,avr-gcc4.9) not working
 */
// typedef float16     sdouble; /* flt16 =  s1+e5+m10=16bits 3.3decdigits */
typedef _Float16     sdouble; /* flt16 =  s1+e5+m10=16bits 3.3decdigits */
#elif (VecType & (1+8)) == 0 /* 32bits = a1+e8+m23  = 7.2decdigits */
typedef float       sdouble;    
#else                        /* 64bits = a1+e11+m52 = 15.9decdigits */
typedef double      sdouble;
/* ToDo: test 16bit fix vs. half-float s1+e5+m10 = 3.3decdigits */
#endif   /* VecType ... == 8 */

#if (VecType & 2) == 2   /* extra accuracy for sums over lot of objects  */
/* gcc                      (default  80bits for x87,    s1+e15+m64) 19dec */
/* gcc -m96bit-long-double  (default  96bits for i386-32) */
/* gcc -m128bit-long-double (default 128bits for x86-64  s1+e15+m112) */
/* gcc -mlong-double-128    (default 128bits for x86-64  s1+e15+m112) 34dec */
typedef long double ldouble;  
#else
typedef double      ldouble;
#endif

#if VecType & 4  /* cplx, ToDo19: check cplx by configure? */
#include <stdlib.h>  /* defines __STDC_IEC_559_COMPLEX__ if cplx available */
/* gcc -dM -E -x c /dev/null | grep -i -e COMPLEX
 * #                  __GCC_IEC_559_COMPLEX 2    # gcc492 ~gcc333-447
 * echo "#include <stdlib.h>" | gcc -dM -E -x c - | grep -i -e COMPLEX
 * #                  __GCC_IEC_559_COMPLEX 2    # gcc492 ~gcc447
 * #                  __STDC_IEC_559_COMPLEX__ 1 # gcc333-492 g++333
 * echo -e "#include <stdlib.h>\n#include <complex.h>"|gcc -dM -E -x c -\
 *   |grep -i -e COMPLEX
 * #                  _GLIBCXX_USE_C99_COMPLEX 1       # gcc333
 * #                  _Complex_I (__extension__ 1.0iF) # gcc333
 * #                  I _Complex_I      # gcc333
 * #                  complex _Complex  # gcc333 (g++333-412=fail)
 * # does not work with g++ or gcc -x c++ !!! g++333-412=fail g++447-492=ok
 * ToDo19: better way to use g++ and (older?) gcc-complex implementation?
 * plz do not use c++ of gcc for complex numbers
 * echo -e "#include <stdlib.h>\n#include <complex>"|gcc -dM -E -x c++ -\
 *   |grep -i -e COMPLEX
 *                    __STDC_IEC_559_COMPLEX__ 1
 * https://en.cppreference.com/w/c/numeric/complex  # C11 2019
 *  #include stdio.h+complex.h+tgmath.h + double complex z1=1.+2*I;
 *           creal(z1), cimag(z1)
 * https://en.cppreference.com/w/cpp/numeric/complex # 
 *  #include iostream+iomanip+complex+cmath + std::complex<double> z1=1.+1i;
 *           z1.imag z1.real
 */
/* gcc/g++492 defines __GCC_IEC_559_COMPLEX=2 (w/wo stdlib) */
/* gcc/g++412/icc/icpc16 + math.h|stdlib.h = __STDC_IEC_559_COMPLEX__=1 */
/* gcc/g++412/icc/icpc16 + complex.h = __STDC_IEC_559_COMPLEX__=1 complex */
/* clang/clang++3.4.2  __STDC_IEC_559_COMPLEX__=1 I=_Complex_I */
#define IFcplx(x) x       /* operations for COMPLEX only */
/* undefined MACROS are defined as 0 in macro expressions */
/* 2018-11 complex.h tested using if 0 && ..., IEC_559 is the default */
/* 2018-11 ToDo: check IEC_559_COMPLEX outside in configure + set switch */
/* 2019-01 test on galileo.gcc333 fails c++ && gcc-complex, but gcc is ok */
/* ??? c++ problems with implic. type conv. if(cplx!=int) a=cplx*scplx
 *   therefore use gcc-implementation for c++ too if possible (2019-01)
 *   who can help? see below + outcomment "|| ... >= ..." lines
 */
#if  defined(__STDC_IEC_559_COMPLEX__) && ((!defined(__cplusplus))\
   || __GNUC__ * 100 +  __GNUC_MINOR__ * 10 >= 440\
   || __ICC >= 1600 || ( __clang_major__ >= 3 && __clang_minor__ >= 4 ))
/* CC="gcc492 -std=c99" (pure gcc4 produces warnings wthout c99)
 *    (__STDC_VERSION__ >= 199901L)
 * g++/gcc -x c++ && _Complex :  gcc-3.3-4.1 = fail, gcc-4.4-4.6-4.9 = ok
 */
/* typedef complex double cplx;
 *    pathcc3 gcc4 g++4 gcc5 g++5 icc16 icpc16 = OK   pathCC3 g++6 = FAIL 
 * typedef _Complex double cplx;
 *    pathCC3 gcc4 g++4 g++44 gcc5 g++5 g++6 icc16 icpc16 = OK 
 */
#define CplxInfo "__STDC_IEC_559_COMPLEX__ implementation"
/* gcc -std=c99 ... __STDC_VERSION__ 199901L _Complex ... */
/* using C99 complex type (faster) 2013-04-03 by A.Honecker */
/* tested: gcc -std=c99  VecType=  2013-04-04 gcc-4.1.2 test=OK */ 
/* complex.h needed by gcc, but g++ needs math.h or stdlib.h instead */
//#include <complex>      /* kautz g++412 wants this, but failes anyhow */
//#include <math.h>       /* clang doc has math.h before complex.h */
#include <complex.h>      /* _Complex_I defined here */
typedef _Complex double    cplx;
#if   (VecType & (1+8)) == 0          /* float for big arrays */
/* ToDo: C99: "double complex"(c99), not "complex double"(gcc) ! */
typedef _Complex float    scplx;
#elif (VecType & (1+8)) == 8          /* half float to store big arrays */
typedef _Complex float16  scplx; /* float16 _Complex = C99 s1+e5+m10=16bits */
#else
typedef _Complex double   scplx; /* default == 1 double for big arrays */
#endif
#if (VecType & 2) == 2          /* long double to store big arrays */
typedef _Complex long double  lcplx;
#else
typedef _Complex double       lcplx;
#endif
#define D(z)   ((double)creal(z))  /* real part of z */
#define Re(z)   creal(z)     /* real part of z */
#define Im(z)   cimag(z)
#define Conj(z) conj(z) 
/* c++11: CMPLX(x, y) = (double complex)(x+_Imaginary_I*y) */
/* _Complex_I, I failes on pathCC3, g++-4.1.2 (sicortex) */
/* _Complex_I ok on gcc-4.1.2, but no _Imaginary_I */
#define C(x,y)     ((x)+(y)*_Complex_I)
#define polar(r,t) ((r) * (cos(t) + _Complex_I*sin(t)))
#define arg(x) (atan2(cimag(x), creal(x)))   /* ToDo: carg(x) ??? */
#define Cabs(x) cabs(x) /* C99 cabs(z)=hypot(a,b)=sqrt(a*a+b*b) glibc2.1 */
#define Norm(x) (creal(x)*creal(x)+cimag(x)*cimag(x))   /* r*r+i*i */
#elif defined(__STDC__) && defined(__GNUC__) && !defined(__cplusplus)
/* gcc-4.1.2 defaults to ISO C90 without __complex__ */
#define CplxInfo "c89+gnu __complex__ implementation"
/* GCC C89/C90 __complex__ float|double|long double (not ISO-C90) */
/* typedef __complex__ double cplx;
 *    pathcc3 pathCC3 gcc4 g++4 gcc5 g++5 icc16 icpc16 = OK
 */
typedef __complex__ double   cplx;
#if   (VecType & (1+8)) == 0          /* float for big arrays */
typedef __complex__ float   scplx;
#elif (VecType & (1+8)) == 8          /* double to store big arrays */
typedef __complex__ float16 scplx;
#else
typedef __complex__ double  scplx;
#endif
#if (VecType & (2)) == 2          /* long double to store big arrays */
typedef __complex__ long double  lcplx;
#else 
typedef __complex__      double  lcplx; 
#endif
#define C(x,y)  ((x)+(y)*1i)
#define D(x)   ((double) __real__ (x))
#define Re(x)   (__real__ (x))
#define Im(x)   (__imag__ (x))
#define Conj(x) (~(x))                 /* ~(x) not on ISO C */
#define arg(x) (atan2(Im(x), Re(x)))   /* ToDo: carg(x) ??? */
#define polar(r,t) ((r) * (cos(t) + 1i*sin(t)))
#define Cabs(x) cabs(x)  /* C99 cabs(z)=hypot(a,b)=sqrt(a*a+b*b) glibc2.1 */
#define Norm(x) (Re(x)*Re(x)+Im(x)*Im(x))   /* r*r+i*i */
/* replace 1 && ... by 0 && ... if complex failes on cpp/c++  2019-01 */
#elif 0 && defined(__cplusplus)  /* using self defined class complex (preC99 c++ compilers) */
/* ??? c++ problems with implic. type conv. if(cplx!=int) a=cplx*scplx */
/* ToDo: https://en.cppreference.com/w/cpp/numeric/complex */
/* ToDo: include <complex>  template <class T>;  class complex; T.imag */
#include <complex>
#define CplxInfo "complex C++ template/class"
typedef std::complex<double>       cplx; /* complex double      (compute) */ 
typedef std::complex<float>       scplx; /* complex float       (storing) */
typedef std::complex<long double> lcplx; /* complex long double (big sums) */
#define C(x,y)  (x+y*1i)
#define D(x)   ((double)x.real)
#define Re(x)   real(x)
#define Im(x)   imag(x)
#define Conj(x) conj(x)
#define Cabs(x) sqrt(norm(x))   /* 2019 no cabs in (older) c++ */
#define Norm(x) norm(x)   /* r*r+i*i */
/* ToDo: may be its better to define I and use macros below */
#elif defined(__cplusplus)  /* using self defined class complex (preC99 c++ compilers) */
/* pathCC3 g++4 g++5 g++6 icpc16 = OK */
/* ToDo: https://en.cppreference.com/w/cpp/numeric/complex */
/* ToDo: include <complex>  template <class T>;  class complex; T.imag */
#if defined(__GNUC__)
#define CplxInfo "slow(?) own complex C++ class (please use gcc, not g++)"
#else
#define CplxInfo "slow(?) own complex C++ class (no gcc available?)"
#endif
#include "complex.h"
typedef class  Complex  cplx; /* complex double      (compute) */ 
typedef class sComplex scplx; /* complex float       (storing) */
typedef class lComplex lcplx; /* complex long double (big sums) */
#define C(x,y)  Complex(x,y)
#define D(x)   ((double)real(x))
#define Re(x)   real(x)
#define Im(x)   imag(x)
#define Conj(x) conj(x)
#define Cabs(x) cabs(x)   /* 2018-11 replace fabs by cabs */
#define Norm(x) norm(x)   /* r*r+i*i */
/* ToDo: may be its better to define I and use macros below */
#else  /* c++ or c99 */
#   error "use c++ OR add -std=c99 !!!"
#endif
#else  /* ! ( VecType & 4 ) == real, ToDo: define I as 0 ??? */
#define CplxInfo "no complex numbers used, all numbers real"
#define IFcplx(x)
typedef  double   cplx;
typedef sdouble  scplx; /* short for huge arrays */
typedef ldouble  lcplx; /* long  for huge summs */
#define C(x,y)  ((double)(x))
#define D(x)    ((double)(x))
#define Re(x)   (x)
#define Im(x)   0.0        /* 2020-11 fix warning 0 (int) */
#define Conj(x) (x)
#define Cabs(x) fabs(x)
#define Norm(x) norm(x)   /* not needed!? */
/* #define Norm(x) fsqr(x) */
#endif

/*
#ifndef CplxInfo
#define CplxInfo ""
#endif
*/
#if !defined(M_d2f) /* VecType & 8 needs special handling, see above */
/*
//#define M_d2f(x) ((scplx)x)
//#define M_f2d(x) ( (cplx)x)
// ToDo: d2f+f2d is obsolete, before removing it, we test a while 2013-04-04
 */
#define M_d2f(x) (x)
#define M_f2d(x) (x)
#endif

#include <limits.h>  /* __WORDSIZE (32,64) __LONG_LONG_MAX__ CHAR_BIT */
/* this values are fixed, but checked for sureness */
#ifndef S_int
#define S_int 4
#endif
#ifndef S_long
/* echo -e "#include <stdlib.h>"|gcc -dM -E -x c -|grep WORDSIZE
 * # __WORDSIZE_COMPAT32 1  gcc447.x86_64 + __x86_64 __S64_TYPE __U64_TYPE
 * # __WORDSIZE 64          gcc447.x86_64
 * # -empty-                gcc333.i386=galileo
 */
#ifdef __WORDSIZE   /* linux */
#define S_long (__WORDSIZE/8)  /* 32 or 64 */
#elif defined(LONG_BIT) /* Tru64 */
#define S_long (LONG_BIT/8)    /* 32 or 64 */
#elif __LONG_MAX__ == 2147483647L
#define S_long (32/8)   /* i386+gcc333 2019-01 */
#else  /* just guessing, check it later, 8: fail on gcc333.i386 */
#define S_long (64/8)
#endif
#endif
/* long long is not ansi and therefore not used */
typedef unsigned long ULong;  /* already defined? -> where? Warnings */

#define Dsize sizeof(sdouble)
#define Lsize sizeof(long)
#define Csize sizeof(scplx)

/* using "\" in __TURBOC__ produces wrong line numbers!	*/
#define FF {}     	/* fflush(stdout) -> setvbuff(stdout,...)	*/

#ifdef sun
#undef sun                      /* no problems with 'SUN's (-Dsun) */
#endif

#include "array_m.h"		/* needs definition of scplx */

/* SLES9 forgets min max if array_m.h follows, cplx=dbl */
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))

#endif /* ifndef _DEFS_H_ */
