/* 
    =>>> look at error.h
    
  Apr09: move \n at line end to avoid buffering on MPI systems
*/

#include <stdlib.h>
#include <stdio.h>
// t100.gpu02 does not like this
#include <errno.h>
#include <unistd.h> /* sleep() gethostname() */
#include "error.h"
#include "block.h" // possible set CONFIG_MPI ?
#ifdef CONFIG_MPI
#if  defined(__cplusplus) && !defined(MPICH_IGNORE_CXX_SEEK)
/* may be including stdio.h before mpi.h will work too */
#define MPICH_IGNORE_CXX_SEEK  /* avoid macro conflicts mpich with stdio.h */
#endif /* __cplusplus */
#include <mpi.h>
#endif

int vvv=3;            /* verbose mode */
int err_fatal=0,      /* public, to inform calling functions */
    err_sum=0;        /* num of all errors, slowing down program */
/* pthread access is uncritical in error case */
extern int mpi_i, mpi_n;     /* CONFIG_MPI */
extern char nodename[21]; /* 2016-04-08 set global for err() see spins.c */             
      
#define FF fflush(stdout)  // may not help in MPI-environments

/* this is a general error routine, which limits the output
 * ToDo: reset by external signal, send a SIGSTOP to itself? (allow SIGCONT)
 * call via ERR(no) -> err_pg_dbg(no,__FILE__,__LINE__)
 *  cost more code size (12-24Byte 6*x86asm), but gives more info
 * ERR(-1)      = reset
 * ERR(0)       = dummy
 * ERR(1)       = debug info file+line
 * ERR(2..8999) = soft error
 * ERR(9000..)  = hard error
 ToDo: mpi: concat multiple messages, async send to node0, recv from0
   node0: send to all nodes async
 */
void err_pg(int no, const int line, const char *file){
  /* avoid thousend of error outputs on massiv parallel computers */
  /* 7&(7-1)=6, 8&(8-1)=0 */
  int is_log_mpi=(((mpi_i&(mpi_i-1))==0)?1:0); /* only task 0-2,4,8,16,... */
  if (no==1 && is_log_mpi) /* mpi_i==2^n only log errors */
    printf("\nDBG %s L%d node %3d\n", file, line, mpi_i);
  /* ToDo: sleep(log(mpi_i)) or Isend/Irecv to sync/reduce parallel err msgs */
  if (no==-1) {
    if (err_fatal && (mpi_i==0 || mpi_i+1==mpi_n)) { /* limit 2 outputs */
      /* err_pg(1,line,file); */ 
      printf("\nreset ERRNO=%d %s L%d",err_fatal,file,line); }
    errno=0; err_fatal=0; }
  if (no>1) {
    if (is_log_mpi) printf("\n");
    if (no != 620 || is_log_mpi) /* 2015-02 OOM=620 is likely on every node */
      printf("ERROR %d doc/error.html tsk%03d %s num=%d errno=%d %s L%d\n",
             no, mpi_i, nodename, err_sum, errno, file, line);FF;
    /* output errno-code, man errno fmtmsg strerror oserror */
    if (errno && mpi_i<3) perror("PERROR"); /* max 3 outputs */
    /* avoid gigant error log-files (full disk) */
    err_sum++;		/* count errors */
    if (err_sum== 100) {if(is_log_mpi)printf(" sleep 2s on error\n");FF;}
    if (err_sum>= 100) sleep(2);
    if (err_sum>=1000) {if(is_log_mpi)printf(" sleep 10m\n");FF;sleep(60*10);}
  }
  if (no>1 && no>err_fatal) err_fatal=no; /* 2017-01-23 fix mpi+ev hang */
#ifdef CONFIG_MPI
  // hangs 2019-07-03 on nu=32 nd=32 N=16 
  //if (no>=9000) printf("DBG: call MPI_Abort\n");
  if (no>=9000) MPI_Abort(MPI_COMM_WORLD,no); /* may hang else */
#else
  //if (no>=9000) printf("DBG: call exit\n");
  if (no>=9000) exit(no);
#endif
  return;
}
